# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

from typing import Dict, Any, Optional,Final
from .enums import PathData, Ukrainization
from .json_data import JsonManager
from .http import AioSession
from ..model import ukrainization_model 


MAIN_LINK: Final[str] = "https://raw.githubusercontent.com/DEViantUA/StarRailCardUA/main/{file}"


class TranslateDataManager:
    
    _data =  Dict[str, Optional[Dict[str, Any]]]
    
    def __init__(self):
        self.data = {
            "stats": None,
            "weapons": None,
            "avatar": None,
            "relict_sets": None,
            "paths": None,
            "element": None,
        }
    
    async def load_translate_data(self):
        for file in Ukrainization:
            self.data[file] = await JsonManager(str(PathData.UKRAINIZATION.value / f'{file}.json')).read()
        TranslateDataManager._data = ukrainization_model.UkrainizationModel(**self.data)
        
    async def update(self):
        for file in Ukrainization:
            data = await AioSession.get(MAIN_LINK.format(file = f"{file}.json"))
            await JsonManager(str(PathData.UKRAINIZATION.value / f'{file}.json')).write(data)
        
    async def check_update(self):
        keys = await JsonManager(str(PathData.UKRAINIZATION.value / "keys.json")).read()
        keys_data = await AioSession.get(MAIN_LINK.format(file = "keys.json"))
        
        if keys_data["key"] == keys["key"]:
            return True
        
        await self.update()
        await JsonManager(str(PathData.UKRAINIZATION.value / "keys.json")).write(keys_data)
        
        
        
            
        
    


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.


import aiohttp
import base64

from io import BytesIO
from .src.model import utils_model
from contextlib import AsyncExitStack


import_magic = False

try:
    import magic
except ImportError:
    import_magic = True
    import imghdr

IMAGE_TYPES = {
    'image/jpeg',
    'image/png',
    'image/gif'
}


async def get_character(lang = "en"):
    url = f"https://hsr.yatta.top/{lang}/archive/avatar"
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            return await response.json()

async def get_name(characrer_id = None, lang = "en"):
    avatar_data = await get_character(lang)
    if avatar_data.get("response") != 200:
        print("An error occurred while receiving data")
        return 
    data = {}
    for key in avatar_data.get("data", {}).get("items", {}):
        if not characrer_id is None:
            if int(key) == int(characrer_id):
                return {key: avatar_data["data"]["items"][key].get("name")}
        data[key] = avatar_data["data"]["items"][key].get("name")

    return data

async def get_character_id(name = None, lang = "en"):
    
    avatar_data = await get_character(lang)
    if avatar_data.get("response") != 200:
        print("An error occurred while receiving data")
        return 
    data = {}
    for key in avatar_data.get("data", {}).get("items", {}):
        if not name is None:
            if name in avatar_data["data"]["items"][key].get("name"):
                return {avatar_data["data"]["items"][key].get("name"): key}
        data[avatar_data["data"]["items"][key].get("name")] = key

    return data



async def image_to_base64(image, format = "png"):
    buffered = BytesIO()
    image.save(buffered, format= format)
    img_str = base64.b64encode(buffered.getvalue())
    
    return img_str

async def get_link_image(img, api_key = None):
    """_summary_

    Args:
        img (base64,str): Image
        api_key (str, optional): The API key was obtained here: https://api.imgbb.com/. Defaults to None.

    Raises:
        TypeError: If api_key = None

    Returns:
        ApiImageLink: A class object containing information about the image uploaded to the server
    """
    if api_key is None:
        raise TypeError(800,'Get the API key on the page: https://api.imgbb.com/ (Click on "Get Api Key")')
    
    async with aiohttp.ClientSession() as session:
        url = 'https://api.imgbb.com/1/upload'
        headers = {
            'Accept': 'application/json'
        }

        par = {
            'key': api_key,
            'image': await image_to_base64(img) if not isinstance(img, str) else img
        }
        async with session.post(url, data=par, headers=headers) as req:
            data = await req.json()
            return utils_model.ApiImageLink(**data)   




def get_pixv_headers():
        return {"referer": "https://www.pixiv.net/"}


async def download_image(session, url, headers=None, allow_redirects=True, use_range=None, offset=0, size=None, **kwargs):
    if headers is None:
        headers = {}

    params = {
        'headers': headers,
        'allow_redirects': allow_redirects,
        **kwargs,
    }

    if use_range is None:
        async with session.head(url, **params) as response:
            if response.headers.get('accept-ranges') == 'bytes':
                use_range = True
            else:
                use_range = False

    if use_range:
        if size is None:
            if offset:
                headers['range'] = f"bytes={offset}-"
        else:
            headers['range'] = f"bytes={offset}-{offset + size - 1}"

    async with session.get(url, headers=headers, allow_redirects=allow_redirects, **kwargs) as response:
        if response.status != 416:
            response.raise_for_status()

            buffer = BytesIO()
            async for data in response.content.iter_any():
                length = len(data)

                if use_range:
                    if size is not None and offset + length > offset + size:
                        buffer.write(data[:offset + size - offset])
                        break
                    else:
                        buffer.write(data)
                else:
                    buffer.write(data)

            return buffer.getvalue()

async def get_mimetype(session, url, size=2048, allow_redirects=True, **kwargs):
    async with session:
        data = await download_image(session, url, size=size, allow_redirects=allow_redirects, **kwargs)
        if import_magic:
            mime_type = imghdr.what(None, h=data)
            if mime_type:
                return mime_type
            else:
                return "application/octet-stream"
        else:
            return magic.Magic(mime=True).from_buffer(data)

async def is_valid(session, url, allow_redirects=True, **kwargs):
    async with session:
        try:
            async with session.head(url, allow_redirects=allow_redirects, **kwargs) as response:
                return response.status == 200
        except aiohttp.ClientError:
            return False
    
async def is_valid_image(url,session=None,allow_redirects=True,strict=False,**kwargs):
    """Check the image from the link or not

    Args:
        url (str): link
        session (session, optional): aiohttp or other session. Defaults to None.
        allow_redirects (bool, optional): Allow redirection. Defaults to True.
        strict (bool, optional): .... . Defaults to False.

    Returns:
        bool: True - if the link leads to an image | False - if the link does not lead to an image
    """
    async with AsyncExitStack() as stack:
        if url is None:
            session, url = url,session
        
        if session is None:
            session = await stack.enter_async_context(aiohttp.ClientSession())
            
        async with session:
            try:
                async with session.head(url, allow_redirects=allow_redirects, **kwargs) as response:
                    if response.status != 200 or not response.headers.get('content-length'):
                        return False

                    content_type = response.headers.get('content-type')

                    if content_type == 'application/octet-stream' or strict:
                        use_range = response.headers.get('accept-ranges') == 'bytes'
                        content_type = await get_mimetype(session, url, use_range=use_range, allow_redirects=allow_redirects, **kwargs)

                    return content_type in IMAGE_TYPES
            except aiohttp.ClientError:
                return False


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.


class RelictScore:
    #TOTAL  SETTING
    RGBA = "RGBA"
    
    #BACKGROUND  SETTING
    font = 25
    bacground_size = (1920,782)
    art_size = (782,782)
    splash_size = (850,782)
    blur_art = 30
    blur_splashart = 25
    opacity_art = 0.5
    opacity_splash = 0.7
    position_line = (693,-7)
    
    #UID SETTING
    font_uid = 10
    position_uid = (7,768)
    
    #NAME SETTING
    font_name_size = 25
    name_with = 262
    name_color = (255, 255, 255, 255)
    name_h_max = 53
    name_h_min = 35
    name_position = 4
    name_size = (288, 103)
    font_name_level = 17
    color_name_level = (255, 255, 255, 255)
    position_name_level = (4, 60)
    position_name_star = (2,85)
    
    #CONSTANT SETTING
    constant_size_background = (403, 63)
    constant_size= (163, 164)
    constant_size_icon = (53,53)
    constant_size_icon_opacity = 0.2
    constant_blur = 3
    constant_icon_position = (5,5)
    
    #RELICT SETTING
    relict_size = (273,134)
    relict_icon_size = (113, 113)
    relict_icon_position = (-6,21)
    relict_main_stat_icon_size = (48, 48)
    relict_main_stat_icon_position = (90,41)
    relict_position_star = (51,27)
    relict_sub_icon_size = (29, 29)
    
    #RELICT_SETS SETTING
    sets_background = (559,56)
    setst_line_size = (559,28)
    setst_name_color = (255, 200, 91, 255)
    setst_count_position = (8,4)
    setst_font = 18
    
    #STATS SETTING
    stat_font_dop = 12
    stat_font = 20
    stat_value_font = 20
    stat_line_size = (273,58)
    stat_name_size = 15
    stat_max_width = 140
    stat_icon_size = (45,45)
    stat_icon_position = (0,6)
    stat_name_position = (46,29)
    stat_x_position = 271
    stat_line_add = False

class Ticket:
    #TOTAL  SETTING
    RGBA = "RGBA"
    
    #LIGHT_CONE SETTING
    lc_bacgrount_size = (337, 448)
    lc_image_size = (298, 410)
    lc_image_position = (19, 23)
    
    lc_frame_position = (32, 29)
    
    #BACKGROUND  SETTING
    font = 25
    bacground_default_color = (30,30,30,255)
    bacground_size = (1924,802)
    background_splash_size = (1341,802)
    art_size = (693,802)
    splash_size = (582,802)
    blur_art = None
    blur_splashart = None
    opacity_art = None
    opacity_splash = None
    position_shadow = (1341,0)
    position_line = (0,0)
    position_art = (1341,0)
    position_splash_art = (1342,0)
    
    #UID SETTING
    font_uid = 10
    position_uid = (1443 ,779)
    
    #NAME SETTING
    font_name_size = 25
    name_with = 262
    name_color = (255, 255, 255, 255)
    name_h_max = 53
    name_h_min = 35
    name_position = 4
    name_size = (288, 103)
    font_name_level = 17
    color_name_level = (255, 255, 255, 255)
    position_name_level = (4, 60)
    position_name_star = (2,85)
    
    #CONSTANT SETTING
    constant_size_background = (403, 63)
    constant_size= (163, 164)
    constant_size_icon = (53,53)
    constant_size_icon_opacity = 0.2
    constant_blur = 3
    constant_icon_position = (5,5)
    
    #RELICT SETTING
    relict_size = (273,134)
    relict_icon_size = (113, 113)
    relict_icon_position = (-6,21)
    relict_main_stat_icon_size = (38, 38)
    relict_main_stat_icon_position = (77,0)
    relict_position_star = (0,20)
    relict_sub_icon_size = (38, 38)
    
    #RELICT_SETS SETTING
    sets_background = (807,545)
    setst_line_size = (394,38)
    setst_name_color = (255, 185, 85, 255)
    setst_count_position = (7,2)
    setst_font = 15
    
    #STATS SETTING
    stat_font_dop = 18
    stat_font = 20
    stat_value_font = 30
    stat_line_size = (434,52)
    stat_name_size = 20
    stat_max_width = 275
    
    stat_y_no_dop = 15
    stat_y_yes_dop = 3
    stat_y_dop = 32
    
    stat_icon_size = (52,52)
    stat_icon_position = (0,0)
    stat_name_position = (59,27)
    stat_x_position = 432
    stat_line_add = False


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

import json
import aiofiles


class JsonManager:
    def __init__(self, file_path):
        self.file_path = file_path

    async def read(self):
        try:
            async with aiofiles.open(self.file_path, mode='r', encoding="utf-8") as file:
                data = await file.read()
                return json.loads(data)
        except FileNotFoundError:
            print(f"File '{self.file_path}' not found.")
        except json.JSONDecodeError:
            print(f"Error decoding JSON file '{self.file_path}'.")
    
    async def write(self, data):
        try:
            async with aiofiles.open(self.file_path, mode='w') as file:
                await file.write(json.dumps(data, indent=4, ensure_ascii=False))
        except Exception as e:
            print(f"Error writing to file '{self.file_path}': {e}")


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.


from pydantic import BaseModel
from typing import Dict, Union

class UkrainizationModel(BaseModel):
    stats: Dict[str, Union[str, dict, int, None]]
    weapons: Dict[str, Union[str, dict, int, None]]
    avatar: Dict[str, Union[str, dict, int, None]]
    relict_sets: Dict[str, Union[str, dict, int, None]]
    paths: Dict[str, Union[str, dict, int, None]]
    element: Dict[str, Union[str, dict, int, None]]


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

from PIL import Image
from io import BytesIO

import re
import json


from .. import cashe, http

_caches = cashe.Cache.get_cache()

async def resize_image(image, scale):
    width, height = image.size
    new_width = int(width * scale)
    new_height = int(height * scale)
    resized_image = image.resize((new_width, new_height))
    return resized_image


async def get_centr_size(size, file_name): #get_centr_honkai_art
    background_image = Image.new('RGBA', size, color=(0, 0, 0, 0))
    foreground_image = file_name.convert("RGBA")

    scale = max(size[0] / foreground_image.size[0], size[1] / foreground_image.size[1])
    foreground_image = foreground_image.resize((int(foreground_image.size[0] * scale), int(foreground_image.size[1] * scale)))

    background_size = background_image.size
    foreground_size = foreground_image.size

    x = background_size[0] // 2 - foreground_size[0] // 2

    if foreground_size[1] > background_size[1]:
        y_offset = max(int(0.3 * (foreground_size[1] - background_size[1])), int(0.5 * (-foreground_size[1])))
        y = -y_offset
    else:
        y = background_size[1] // 2 - foreground_size[1] // 2

    background_image.alpha_composite(foreground_image, (x, y))

    return background_image


async def get_centr_scale(size, file_name):
    
    background_image = Image.new('RGBA', size, color=(0, 0, 0, 0))
    foreground_image = file_name.convert("RGBA")
    
    scale = 0.65
    foreground_image = await resize_image(foreground_image, scale)

    background_size = background_image.size
    foreground_size = foreground_image.size

    x = background_size[0] // 2 - foreground_size[0] // 2
    y = background_size[1] // 2 - foreground_size[1] // 2

    background_image.alpha_composite(foreground_image, (x, y))

    return background_image


async def get_dowload_img(link,size = None, thumbnail_size = None, gif = False):
    cache_key = json.dumps((link, size, thumbnail_size), sort_keys=True)
        
    if not gif:
        if cache_key in _caches:
            return _caches[cache_key]
    
    headers_p = None
    
    if "pximg" in link:
        headers_p = {
            "referer": "https://www.pixiv.net/",
        }
    try:
        image = await http.AioSession.get(link,headers=headers_p, response_format= "bytes")
    except:
        raise
    
    if gif:
        return image
        
    image = Image.open(BytesIO(image)).convert("RGBA")
    if size:
        image = image.resize(size)
        _caches[cache_key] = image
        return image
    elif thumbnail_size:
        image.thumbnail(thumbnail_size)
        _caches[cache_key] = image
        return image
    else:
        _caches[cache_key] = image
        return image

async def crop_image(img):
    width, height = img.size
    target_pixel_x = 275
    target_pixel_y = height // 2
    crop_size = 8
    left = max(0, target_pixel_x - crop_size // 2)
    right = min(width, target_pixel_x + crop_size // 2 + 1)
    cropped_img = img.crop((left, 0, right, height))
    
    return cropped_img

async def get_user_image(img):
    if type(img) != str:
        img = img
    elif type(img) == str:
        linkImg = re.search("(?P<url>https?://[^\s]+)", img)
        if linkImg:
            try:
                if "gif" in linkImg.group():
                    img = await get_dowload_img(linkImg.group(), gif = True)
                    return img
                else:
                    img = await get_dowload_img(linkImg.group())
            except:
                return None
        else:
            img = Image.open(img)
    else:
        return None
    return img.convert("RGBA")



# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

import cachetools

class Cache:
    _cache = None

    @classmethod
    def get_cache(cls, maxsize: int = 100, ttl: int = 60) -> cachetools.TTLCache:
        """
        Returns the cache instance, creating it if necessary.

        Args:
            maxsize (int, optional): Maximum number of items the cache can hold. Defaults to 100.
            ttl (int, optional): Time-to-live for cached items in seconds. Defaults to 60.

        Returns:
            cachetools.TTLCache: The cache instance.
        """
        if cls._cache is None:
            cls._cache = cachetools.TTLCache(maxsize=maxsize, ttl=ttl)
        return cls._cache

    @classmethod
    def clear_cache(cls):
        """
        Clears the cache.
        """
        if cls._cache is not None:
            cls._cache.clear()
            cls._cache = None

    @classmethod
    def setting(cls, maxsize: int = 100, ttl: int = 60) -> cachetools.TTLCache:
        """
        Setting the cache.
        """
        cls._cache = cachetools.TTLCache(maxsize=maxsize, ttl=ttl)
    
    @classmethod
    def get(cls, key):
        """
        Retrieves a value from the cache by key.

        Args:
            key: The key of the value to retrieve.

        Returns:
            Any: The value associated with the key, or None if the key is not found in the cache.
        """
        if cls._cache is not None:
            return cls._cache.get(key)

    @classmethod
    def set(cls, key, value):
        """
        Sets a value in the cache with the specified key.

        Args:
            key: The key to set.
            value: The value to associate with the key.
        """
        if cls._cache is not None:
            cls._cache[key] = value

    @classmethod
    def delete(cls, key):
        """
        Deletes a value from the cache by key.

        Args:
            key: The key of the value to delete.
        """
        if cls._cache is not None and key in cls._cache:
            del cls._cache[key]

    @classmethod
    def contains(cls, key):
        """
        Checks if the cache contains a value with the specified key.

        Args:
            key: The key to check.

        Returns:
            bool: True if the key is found in the cache, otherwise False.
        """
        if cls._cache is not None:
            return key in cls._cache
        return False



# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

from enum import  StrEnum, Enum
from pathlib import Path
    
class Ukrainization(StrEnum):
    STATS = "stats"
    WEAPON = "weapons"
    AVATAR = "avatar"
    RELICT_SETS = "relict_sets"
    PATH = "paths"
    ELEMENT = "element"
    
    @classmethod
    def __iter__(cls):
        return iter(file.value for file in cls)
    
class PathData(Enum):
    UKRAINIZATION = Path(__file__).parent.parent / "data"
    
class Style(Enum):
    RELICT_SCORE = 1
    TICKET = 2
    ENKA = 3


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.


from pydantic import BaseModel
from PIL import Image
from typing import List ,Optional,Final, Union
from moviepy.editor import ImageSequenceClip
import numpy as np

from ..tools.ukrainization import TranslateDataManager
from ..tools.http import AioSession

UA_LANG = False
MAIN_LINK: Final[str] = "https://raw.githubusercontent.com/Mar-7th/StarRailRes/master/{icon}"

class MemoryInfo(BaseModel):
    level: Optional[int]
    chaos_id: Optional[int]
    chaos_level: Optional[int]
    
class SpaceInfo(BaseModel):
    memory_data: Optional[MemoryInfo]
    universe_level: Optional[int]
    light_cone_count: Optional[int]
    avatar_count: Optional[int]
    achievement_count: Optional[int]

class Avatar(BaseModel):
    id: str
    name: str
    icon: str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.icon = MAIN_LINK.format(icon = self.icon)
    
        if UA_LANG:
            self.name = TranslateDataManager._data.avatar.get(self.id, self.name)

class Player(BaseModel):
    uid: Optional[str]
    nickname: Optional[str]
    level: int
    world_level: int
    friend_count: int
    avatar: Avatar
    signature: Optional[str]
    is_display: bool
    space_info: Optional[SpaceInfo]

class Card(BaseModel):
    id: int
    name: Optional[str]
    rarity: int
    card: Union[Image.Image,list]
    animation: bool
    size: Optional[tuple]
    color: Optional[tuple]
    class Config:
        arbitrary_types_allowed = True
    
    async def get_info(self, lang = "en"):
        await AioSession.get_session()
        url = f"https://api.yatta.top/hsr/v2/{lang}/avatar/{self.id}"
        data = await AioSession.get(url, response_format = "json")
        data["data"]["icon"] = {"icon": f'https://api.yatta.top/hsr/assets/UI/avatar/medium/{data["data"]["icon"]}.png', 
                                "splash": f'https://api.yatta.top/hsr/assets/UI/avatar/large/{data["data"]["icon"]}.png',
                                "avatar": f'https://raw.githubusercontent.com/Mar-7th/StarRailRes/master/icon/avatar/{data["data"]["icon"]}.png'
        }        
        await AioSession.close_session()
        
        return data["data"]
    
    def show(self):
        if self.animation:
            self.card[0].show()
        else:
            self.card.show()
    
    def save_gif(self, method = "pillow", name_file = None, format = "gif", resize = None):
        if name_file is None:
            name_file = f"{self.id}_{self.rarity}"
        if isinstance(self.card, list):
            if not format in ["gif", "webp"]:
                format = "gif"
            if format == "webp":
                method = "pillow"
            
            if method.lower() == "pillow":
                self.card[0].save(f"{name_file}.{format}", format = format.capitalize(), save_all=True, append_images=self.card[1:], optimize=True, quality=100, duration=100, loop=0)
            else:
                if resize is None:
                    self.card = [np.array(frame) for frame in self.card]
                else:
                    self.card = [np.array(frame.resize(resize)) for frame in self.card]
                clip = ImageSequenceClip(self.card, fps=10)
                clip.write_gif(f"{name_file}.{format}", fps=10, loop=True, program='ffmpeg', fuzz=0, opt='OptimizeTransparency', logger = None )
        else:
            self.card.save(f"{name_file}.png")
        
class Setting(BaseModel):
    uid: int
    lang: Optional[str]
    hide_uid: bool
    save: bool
    force_update: bool
    style: int
    
class StarRail(BaseModel):
    settings: Setting
    player: Player
    card: Union[List[Card], Image.Image]
    character_id: list
    character_name: list
    
    class Config:
        arbitrary_types_allowed = True
    
    def __str__(self):
        return f"card={self.card} character_name = {self.character_name} character_id={self.character_id}"
    
    def get_charter(self, setting = False, name = False):
        if setting:
            card_ids = [str(card.id) for card in self.card]

            if name:
                return {name: id for id, name in zip(self.character_id, self.character_name) if id in card_ids}
            return {id: name for id, name in zip(self.character_id, self.character_name) if id in card_ids}
        
        if name:
            return {name: id for id, name in zip(self.character_id, self.character_name)}
        
        return {id: name for id, name in zip(self.character_id, self.character_name)}


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.


def get_relic_score_color(score):
    if score < 11:
        return (172, 126, 255, 255)
    elif 11 <= score < 17:
        return (172, 126, 255, 255)
    elif 17 <= score < 21:
        return (126, 175, 255, 255)
    elif 21 <= score < 29:
        return (126, 255, 190, 255)
    elif 29 <= score < 35:
        return (187, 255, 126, 255)
    elif 35 <= score < 41:
        return (255, 168, 126, 255)
    elif 41 <= score < 47:
        return (255, 126, 126, 255)
    elif 47 <= score:
        return (255, 126, 126, 255)
    

def get_total_score_color(score):
    if score < 11:
        return (172, 126, 255, 255)
    elif 11 <= score < 17:
        return (172, 126, 255, 255)
    elif 17 <= score < 21:
        return (126, 175, 255, 255)
    elif 21 <= score < 29:
        return (126, 255, 190, 255)
    elif 29 <= score < 35:
        return (187, 255, 126, 255)
    elif 35 <= score < 41:
        return (255, 168, 126, 255)
    elif 41 <= score < 47:
        return (255, 126, 126, 255)
    elif 47 <= score:
        return (255, 126, 126, 255)
    
def get_relic_score_text(score):
    if score < 11:
        return "D"
    elif 11 <= score < 17:
        return "C"
    elif 17 <= score < 21:
        return "B"
    elif 21 <= score < 29:
        return "A"
    elif 29 <= score < 35:
        return "S"
    elif 35 <= score < 41:
        return "SS"
    elif 41 <= score < 47:
        return "SSS"
    elif 47 <= score:
        return "SSS"

def get_relic_full_score_text(score):
    if score < 60:
        return "D"
    elif 60 <= score < 100:
        return "C"
    elif 100 <= score < 140:
        return "B"
    elif 140 <= score < 180:
        return "A"
    elif 180 <= score < 210:
        return "S"
    elif 210 <= score < 250:
        return "SS"
    elif 280 <= score:
        return "SSS"



# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

import asyncio
import io
from PIL import ImageDraw,Image,ImageFilter,ImageChops, ImageSequence

from ..tools import git, options
from ..tools.calculator import stats
from ..model.style import RelictScore

from ..tools import pill

_of = git.ImageCache()

_DEFAULT_SCORE = {'count': 0, 
                  'rolls': {}, 
                  'rank': {'name': 'N/A', 
                           'color': (255, 255, 255, 255)
                    }
}

async def get_cone_frame(x):
    if x == 5:
        return await _of.light_cone_frame_five
    elif x == 4:
        return await _of.light_cone_frame_four
    else:
        return await _of.light_cone_frame_three

class Creat:
    def __init__(self, data, lang, art, hide_uid, uid, seeleland,remove_logo) -> None:
        self.remove_logo = remove_logo
        self.data = data
        self.lang = lang
        self.art = art #Image.open("modified_exampl.gif")#art
        self.hide_uid = hide_uid
        self.uid = uid
        
        self.gif = False
        self.GIFT_BG = []
        
        self.seeleland = seeleland
        self.totall_eff = 0
        self.element_color = self.data.element.color.rgba
    
    async def creat_bacground(self):
        self.background = Image.new(RelictScore.RGBA, RelictScore.bacground_size, (0,0,0,0))
        background_dark = Image.new(RelictScore.RGBA, RelictScore.bacground_size, (0,0,0,100))
        background_art = Image.new(RelictScore.RGBA, RelictScore.bacground_size, (0,0,0,0))
        background_blur = Image.new(RelictScore.RGBA, RelictScore.bacground_size, (0,0,0,0))
        lines = await _of.background_line
        logo = await _of.LOGO_GIT
        shadow = await _of.background_shadow
        maska_art = await _of.background_maska_art
        background_overlay = await _of.background_overlay
        background_maska_blur = await _of.background_maska_blur
        background_default = await _of.background_default
                
        if self.art:
            user_image = await pill.get_centr_size(RelictScore.art_size, self.art)
            bg = await pill.GradientGenerator(user_image).generate(1,RelictScore.bacground_size[1])       
            bg = bg.resize(RelictScore.bacground_size)
            user_image_op = await pill.apply_opacity(user_image, opacity = RelictScore.opacity_art)

            background_blur.alpha_composite(bg.convert("RGBA"))
            background_blur.alpha_composite(user_image_op)
            
            background_blur = background_blur.filter(ImageFilter.GaussianBlur(RelictScore.blur_art))
            line,self.element_color = await pill.recolor_image(lines, self.element_color[:3], light = True)            
        else:
            bg = background_default.copy() 
            user_image = await pill.get_centr_scale(RelictScore.splash_size, await pill.get_dowload_img(self.data.portrait))
            user_image_op = await pill.apply_opacity(user_image, opacity = RelictScore.opacity_splash)
            background_blur.alpha_composite(bg.convert("RGBA"))
            background_blur.alpha_composite(user_image_op)
            background_blur = background_blur.filter(ImageFilter.GaussianBlur(RelictScore.blur_splashart))
            line = await pill.recolor_image(lines, self.element_color[:3])
            
        background_blur.alpha_composite(background_dark)
        self.background.paste(background_blur,(0,0), background_maska_blur.convert("L"))
        if self.art:
            self.background = ImageChops.soft_light(self.background, background_overlay)
                
        background_art.alpha_composite(bg.convert("RGBA"))
        background_art.alpha_composite(user_image)
        self.background.paste(background_art,(0,0),maska_art.convert("L"))
        self.background.alpha_composite(shadow)
        self.background.alpha_composite(line, (RelictScore.position_line))
        if not self.remove_logo:
            self.background.alpha_composite(logo)
        
        if not self.hide_uid:
            uid = Image.new(RelictScore.RGBA, RelictScore.bacground_size, (0,0,0,0))
            d = ImageDraw.Draw(uid)
            
            font_10 = await pill.get_font(RelictScore.font_uid)
            d.text((0,0), f"UID: {self.uid}", font=font_10, fill=(255, 255, 255, 75))
            self.background.alpha_composite(uid,RelictScore.position_uid)
    
    async def creat_light_cone(self):
        self.background_light_cones = Image.new(RelictScore.RGBA, (330, 190), (0, 0, 0, 0))
        if self.data.light_cone is None:
            return
        
        background = Image.new(RelictScore.RGBA, (139, 190), (0, 0, 0, 0))
        image = await pill.get_dowload_img(self.data.light_cone.portrait, size=(118, 169))
        background.alpha_composite(image,(9,9))
        light_cone_frame_stars = await get_cone_frame(self.data.light_cone.rarity)
        background.alpha_composite(light_cone_frame_stars)
        light_cone_frame = await _of.light_cone_frame
        background.alpha_composite(light_cone_frame)
        
        self.background_light_cones.alpha_composite(background)
        
        background = await _of.light_cone_stats
        background = background.copy()
        
        d = ImageDraw.Draw(background)
        
        font_18 = await pill.get_font(18)
        d.text((28, 4), self.data.light_cone.attributes[0].display, font=font_18, fill=(255, 255, 255, 255))
        d.text((28, 30), self.data.light_cone.attributes[1].display, font=font_18, fill=(255, 255, 255, 255))
        d.text((28, 56), self.data.light_cone.attributes[2].display, font=font_18, fill=(255, 255, 255, 255))
        
        self.background_light_cones.alpha_composite(background,(140,100))
        
        background = await _of.light_cone_ups
        background = background.copy()
        
        d = ImageDraw.Draw(background)
        font_12 = await pill.get_font(12)
        up = options.ups(self.data.light_cone.rank)
        x = int(font_12.getlength(str(up))/2)
        d.text((10-x, 4), up, font= font_12, fill=(255, 217, 144, 255))
        
        self.background_light_cones.alpha_composite(background,(140,68))
        d = ImageDraw.Draw(self.background_light_cones)
        d.text((165, 69), f"{self.lang.lvl}: {self.data.light_cone.level}/{options.max_lvl(self.data.light_cone.promotion)}", font=font_18, fill=(255, 255, 255, 255))
        
        names = await pill.create_image_with_text(self.data.light_cone.name, 18, max_width=180, color=(255, 255, 255, 255))
        line = Image.new("RGBA", (1,48), options.color_lc_line.get(str(self.data.light_cone.rarity), (150, 202, 255, 255)))
        
        
        self.background_light_cones.alpha_composite(line,(140,9))
        self.background_light_cones.alpha_composite(names,(146,int(34-names.size[1]/2)))   
    
    async def creat_stats(self):
        self.background_stats = Image.new(RelictScore.RGBA, (563, 290), (0, 0, 0, 0))

        combined_attributes = {}
        dop = {}

        for attribute in self.data.attributes + self.data.additions:
            field = attribute.field
            if field in combined_attributes:
                combined_attributes[field].value += attribute.value
                dop[field]["dop"] = f"+{attribute.display}"
            else:
                dop[field] = {"main": attribute.display, "dop": 0}
                combined_attributes[field] = attribute

        dop = {key: value for key, value in dop.items() if value['dop'] != 0}
        
        xp,y = 0,0

        async for i, background in pill.creat_stats(combined_attributes, dop, 
                      RelictScore.stat_font_dop,RelictScore.stat_font,
                      RelictScore.stat_line_size, RelictScore.stat_name_size, RelictScore.stat_max_width, RelictScore.stat_icon_size,
                      RelictScore.stat_icon_position, RelictScore.stat_name_position, RelictScore.stat_x_position,
                      self.element_color, RelictScore.stat_value_font):
            self.background_stats.alpha_composite(background,(xp,y))
            y += 45
            if i == 5:
                xp = 290
                y = 0
 
    async def creat_name(self):
        self.background_name = Image.new(RelictScore.RGBA, RelictScore.name_size, (0, 0, 0, 0))
        d = ImageDraw.Draw(self.background_name)
        names = await pill.create_image_with_text(self.data.name, RelictScore.font_name_size , max_width=RelictScore.name_with, color= RelictScore.name_color)
        if names.size[1] <= RelictScore.name_h_max:
            self.background_name.alpha_composite(names,(RelictScore.name_position,int(RelictScore.name_h_max-names.size[1]/2)))
        else:
            self.background_name.alpha_composite(names,(RelictScore.name_position,int(RelictScore.name_h_min-names.size[1]/2)))
        
        font_17 = await pill.get_font(RelictScore.font_name_level)
        max_level = options.max_lvl(self.data.promotion)
        level = f"{self.lang.lvl}: {self.data.level}/{max_level}"
        d.text(RelictScore.position_name_level, level, font=font_17, fill= RelictScore.color_name_level)
        starts = await options.get_stars(self.data.rarity)
        self.background_name.alpha_composite(starts,RelictScore.position_name_star)
       
    async def creat_constant(self):
        self.background_skills = Image.new(RelictScore.RGBA, RelictScore.constant_size_background, (0, 0, 0, 0))
        x = 0
        rank = self.data.rank
        for key in self.data.rank_icons[:rank]:
            bg = Image.new(RelictScore.RGBA, RelictScore.constant_size, (0, 0, 0, 0))
            icon = await pill.get_dowload_img(key, size=RelictScore.constant_size_icon)
            icon_blur = icon.filter(ImageFilter.GaussianBlur(RelictScore.constant_blur))
            icon_blur = await pill.recolor_image(icon_blur, self.element_color[:3])
            bg.alpha_composite(icon_blur,RelictScore.constant_icon_position)
            bg.alpha_composite(icon,RelictScore.constant_icon_position)
            self.background_skills.alpha_composite(bg,(x,0))
            x += 68

        for key in self.data.rank_icons[rank:]:
            bg = Image.new(RelictScore.RGBA, RelictScore.constant_size, (0, 0, 0, 0))
            icon = await pill.get_dowload_img(key, size= RelictScore.constant_size_icon)
            icon = await pill.apply_opacity(icon, opacity=RelictScore.constant_size_icon_opacity)
            bg.alpha_composite(icon,RelictScore.constant_icon_position)
            self.background_skills.alpha_composite(bg,(x,0))
            x += 68
     
    async def creat_relict_sets(self):
        rel_set = options.calculator_relict_sets(self.data.relic_sets)

        self.background_sets = Image.new(RelictScore.RGBA, RelictScore.sets_background, (0,0,0,0))
        
        font = await pill.get_font(RelictScore.setst_font)
        
        line_items = []
        i = 0
        for key in rel_set:
            sets = rel_set[key]
            holst_line = Image.new(RelictScore.RGBA, RelictScore.setst_line_size, (0,0,0,0))
            background_count = await _of.relict_count_sets
            background_count = background_count.copy()
            d = ImageDraw.Draw(background_count)
            d.text(RelictScore.setst_count_position, str(sets["num"]), font=font, fill= RelictScore.setst_name_color)
            d = ImageDraw.Draw(holst_line)
            sets_name_font,size = await pill.get_text_size_frame(sets["name"],15,492)
            if key[:1] == "1":
                holst_line.alpha_composite(background_count)
                d.text((36, 4), sets["name"], font=sets_name_font, fill=RelictScore.setst_name_color)
                line_items.append({"setap": i, "line": holst_line})
                i += 1
            else:
                holst_line.alpha_composite(background_count,(532,0))
                d.text((int(521-size), 4), sets["name"], font=sets_name_font, fill=RelictScore.setst_name_color)
                line_items.append({"setap": 1, "line": holst_line})
                
        for key in line_items:
            if key["setap"] == 1:
                self.background_sets.alpha_composite(key["line"],(0,29))
            elif key["setap"] == 2:
                self.background_sets.alpha_composite(key["line"],(0,29))
            else:
                self.background_sets.alpha_composite(key["line"],(0,0))
    
    async def creat_relict(self,relict):
        background_main = Image.new(RelictScore.RGBA,RelictScore.relict_size, (0,0,0,0))
        background = Image.new(RelictScore.RGBA,RelictScore.relict_size, (0,0,0,0))
        background_image = Image.new("RGBA",RelictScore.relict_size, (0,0,0,0))
        relict_frame = await _of.relict_frame
        relict_maska = await _of.relict_maska
        relict_score_frame = await _of.relict_score_frame
        relict_line = await _of.relict_line
        
        background_main.alpha_composite(relict_frame)
        
        image = await pill.get_dowload_img(relict.icon, size= RelictScore.relict_icon_size)
        background_image.alpha_composite(image,RelictScore.relict_icon_position)
        background.paste(background_image,(0,0),relict_maska.convert("L"))
        
        background_main.alpha_composite(relict_score_frame,(8,0))
        background_main.alpha_composite(background)
        background_main.alpha_composite(relict_line)
        score = self.score_info["score"].get(str(relict.id),_DEFAULT_SCORE)
        
        main_stat_icon = await pill.get_dowload_img(relict.main_affix.icon, size= RelictScore.relict_main_stat_icon_size)
        color = options.color_element.get(relict.main_affix.type, None)
        if not color is None:
            main_stat_icon = await pill.recolor_image(main_stat_icon, color[:3])
        stars = await options.get_stars(relict.rarity, type = 1)
        
        background_main.alpha_composite(stars,RelictScore.relict_position_star)
        background_main.alpha_composite(main_stat_icon,RelictScore.relict_main_stat_icon_position)
        
        font_22 = await pill.get_font(22)
        font_14 = await pill.get_font(14)
        font_18 = await pill.get_font(18)
        
        d = ImageDraw.Draw(background_main)
        x = 135 - int(font_22.getlength(relict.main_affix.display))
        d.text((x, 86), relict.main_affix.display, font= font_22, fill= (255,255,255,255))
        
        x = 121 - int(font_22.getlength(f"+{relict.level}")/2)
        d.text((x, 110), f"+{relict.level}", font= font_14, fill= self.element_color)
        
        y_icon = 21
        y_roll = 30
        y_text = 27
                
        eff = 0
        
        for k in relict.sub_affix:
            icon = await pill.get_dowload_img(k.icon, size=RelictScore.relict_sub_icon_size)
            background_main.alpha_composite(icon,(142,y_icon))
            scoreR = score["rolls"].get(k.type,0)
            if k.type in self.score_info["bad"]:
                d.text((173, y_text), str(k.display), font=font_18, fill= (255,255,255,255))
            else:
                eff += 1
                d.text((173, y_text), str(k.display), font=font_18, fill= options.color_scoreR.get(scoreR,(255,255,255,255)))
            d.text((232, y_roll), f"+{scoreR}", font=font_18, fill= options.color_scoreR.get(scoreR,(255,255,255,255)))
            y_icon += 27
            y_text += 28
            y_roll += 27
            
        d.text((22, 4), "Score:", font=font_14, fill = (255,255,255,255))
        d.text((70, 4), str(score["count"]), font=font_14, fill = score["rank"]["color"])
        
        d.text((105, 4), "Rank:", font=font_14, fill = (255,255,255,255))
        d.text((149, 4), str(score["rank"]["name"]), font=font_14, fill = score["rank"]["color"])
        
        d.text((187, 4), "Eff Stat:", font=font_14, fill = (255,255,255,255))
        d.text((247, 4), str(eff), font=font_14, fill = options.color_scoreR.get(eff,(255,255,255,255)))
        
        self.totall_eff += eff
        
        line_f = await _of.relict_frame_line
        line_f = line_f.copy()
        line_f = await pill.recolor_image(line_f, self.element_color[:3])
        background_main.alpha_composite(line_f,(0,21))
        
        return {"position": str(relict.id)[-1:] , "img": background_main}
    
    async def get_path(self):
        self.background_path = await pill.creat_path(self.data)
    
    async def get_score(self):
        self.score_info = await stats.Calculator(self.data).start()
        
    async def build_relict(self):
        self.background_relict = Image.new(RelictScore.RGBA, (1131,297), (0,0,0,0))
        none_relict = await _of.none_relict
        position = {
            "1": (0,163),
            "2": (286,163),
            "3": (572,163),
            "4": (858,163),
            "5": (572,0),
            "6": (858,0),
        }
        
        map = {"1": None, "2": None, "3": None, "4": None, "5": None, "6": None}
        
        for key in self.relict:
            self.background_relict.alpha_composite(key["img"], position.get(key["position"]))
            map[key["position"]] = 0
        
        for key in map:
            if map[key] is None:
                self.background_relict.alpha_composite(none_relict, position.get(key))

        self.background_relict.alpha_composite(self.background_score,(0,112))
        self.background_relict.alpha_composite(self.background_sets,(0,45))
    
    async def creat_seeleland(self):
        self.seelelen = Image.new("RGBA",(1,1), (0,0,0,0))
        
        if self.seeleland:
            self.seelelen = await _of.seeleland_v2
            self.seelelen = self.seelelen.copy()
            font = await pill.get_font(15)
            data = await options.get_seeleland(self.uid, self.data.id)
            if data is None:
                self.seelelen = Image.new("RGBA",(1,1), (0,0,0,0))
                return None
            if not "rank" in data:
                self.seelelen = Image.new("RGBA",(1,1), (0,0,0,0))
                return None
            draw = ImageDraw.Draw(self.seelelen)
            draw.text((143, 7), str(data["sc"]), font=font, fill=(255, 255, 255, 255))
            draw.text((136, 24), f'{data["rank"][:-2]}..', font=font, fill=(255, 255, 255, 255))
            draw.text((127, 42), data["percrank"].replace("top ", ""), font=font, fill=(255, 255, 255, 255))
    
    async def creat_score_total(self):
        self.background_score = Image.new("RGBA",(559,43), (0,0,0,0))
        
        sclor_bg = await _of.relict_backgroundl_score_line
        sclor_bg = sclor_bg.copy()
        sclor_bg_color = await _of.relict_full_score_line
        sclor_bg_color = sclor_bg_color.copy()
        sclor_bg_color = await pill.recolor_image(sclor_bg_color, self.element_color[:3])
        
        result_percentage = (self.score_info["total_score"]['count'] / 260.53) * 100
        pixels_to_fill = int((result_percentage / 100)*599)
        if pixels_to_fill <= 0:
            pixels_to_fill = 1
        sclor_bg.alpha_composite(sclor_bg_color.resize((pixels_to_fill,16)))
        
        font_18 = await pill.get_font(18)
        
        d = ImageDraw.Draw(sclor_bg)
        x = 280 - int(font_18.getlength(f"{round(result_percentage,1)}%")/2)
        frame = await pill.get_colors(await pill.crop_image(sclor_bg), 15, common=True, radius=5, quality=800) 
        ll = await pill.light_level(frame)
        if ll > 0.6 or frame == (51, 51, 51):
            d.text((x, -1), f"{round(result_percentage,1)}%", font=font_18, fill = (0,0,0,166))
        else:
            d.text((x, -1), f"{round(result_percentage,1)}%", font=font_18, fill = (255,255,255,255))
        
        self.background_score.alpha_composite(sclor_bg,(0,27))
        
        font_21 = await pill.get_font(21)
        d = ImageDraw.Draw(self.background_score)
        
        d.text((249, -2), "Score:", font=font_21, fill = (255,255,255,255))
        d.text((320, -2), str(self.score_info["total_score"]['count']), font=font_21, fill = self.score_info["total_score"]["rank"]["color"])
        
        d.text((17, -2), "Summary Rank:", font=font_21, fill = (255,255,255,255))
        d.text((189, -2), str(self.score_info["total_score"]["rank"]["name"]), font=font_21, fill = self.score_info["total_score"]["rank"]["color"])
        
        d.text((393, -2), "Eff Stat:", font=font_21, fill = (255,255,255,255))
        d.text((497, -2), str(self.totall_eff), font=font_21, fill = options.color_scoreR.get(self.totall_eff,(255,255,255,255)))   
    
    async def build(self):
        bg = Image.new(RelictScore.RGBA, RelictScore.bacground_size, (0,0,0,0))
        bg.alpha_composite(self.background_light_cones,(782,16))
        bg.alpha_composite(self.background_stats,(781,220))
        bg.alpha_composite(self.background_name,(6,654))
        bg.alpha_composite(self.background_skills.resize((320,47)),(1033,134))
        bg.alpha_composite(self.background_path.resize((488,428)),(1409,16))
        bg.alpha_composite(self.background_relict,(781,473))
        bg.alpha_composite(self.seelelen,(1120,35))
        
        if self.gif:
            self.background = []
            for key in self.GIFT_BG:
                key.alpha_composite(bg)
                self.background.append(key.convert("RGB"))
        else:
            self.background.alpha_composite(bg)
                    
    async def start(self):   
        _of.set_mapping(1)
        
        if self.art:
            if "gif" in self.art:
                self.gif = True
            self.art = await pill.get_user_image(self.art)
            
        if self.gif:    
            n = 2           
            frame_count = 0
            color = True
            with io.BytesIO(self.art) as f:
                self.art = Image.open(f)
                for frame in ImageSequence.Iterator(self.art):
                    if color:
                        self.element_color = await pill.get_colors(frame.convert("RGBA"), 15, common=True, radius=5, quality=800)
                        color = False
                    if frame_count >= 50:
                        break
                    if frame_count % n != 0:
                        self.art = frame.convert("RGBA")
                        await self.creat_bacground()
                        self.GIFT_BG.append(self.background.copy())
                    frame_count += 1            
        else:
            if self.art:
                self.element_color = await pill.get_colors(self.art, 15, common=True, radius=5, quality=800)
            
            await self.creat_bacground()
            
        await asyncio.gather(
            self.creat_light_cone(),
            self.creat_stats(),
            self.creat_name(),
            self.creat_constant(),
            self.creat_relict_sets(),
            self.get_score(),
            self.get_path(),
            self.creat_seeleland()
        )
        relic_tasks = [self.creat_relict(key) for key in self.data.relics]
        self.relict = await asyncio.gather(*relic_tasks)
        
        await self.creat_score_total()
        
        await self.build_relict()
        await self.build()
        
        data = {
            "id": self.data.id,
            "animation":  self.gif,
            "name": self.data.name,
            "rarity": self.data.rarity,
            "card": self.background,
            "size": RelictScore.bacground_size
        }
        
        return data


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

import asyncio

from .src.tools import http, ukrainization, options, translator, cashe, git
from .src.generator import style_relict_score, style_ticket, style_profile_phone
from .src.api import api
from .src.model import StarRailCard


class Card:
    def __init__(self, lang = "en", character_art = None, character_id = None, seeleland = False,
                 user_font = None, save = False, asset_save = False, remove_logo = False,
                 cashe = {"maxsize": 150, "ttl": 300}):
        """Main class for generating cards

        Args:
            lang (str, optional): Language in which generation will take place, supported languages: cht, cn, de, en, es, fr, id, jp, kr, pt, ru, th, vi, ua . Defaults to "en".
            character_art (dict, optional): A dictionary that contains a key - character id and link value (If you want to pass several images for 1 character, you can use list in the values). Defaults to None.
            character_id (str, optional): List character ids separated by commas. Defaults to None.
            seeleland (bool, optional): Enable Seeleland statistics. Defaults to False.
            user_font (srt, optional): Font path or font name. Defaults to None.
            save (bool, optional): Whether to save images or not (Does not work for animated cards). Defaults to False.
            asset_save (bool, optional): Save assets to the device so that in subsequent calls you do not have to download them, but open them from the device. Defaults to False.
            remove_logo (bool, optional): Remove GItHub logo. Defaults to False.
            cashe (dict, optional): Set your cache settings. Defaults to {"maxsize": 150, "ttl": 300}.
        """
        self.session = None
        self.lang = lang
        self.character_art  = character_art
        self.character_id  = character_id
        self.seeleland = seeleland
        self.user_font = user_font
        self.save = save
        self.asset_save = asset_save
        self.remove_logo = remove_logo
        self.cashe = cashe
        
    async def __aenter__(self):
        cashe.Cache.get_cache(maxsize = self.cashe.get("maxsize", 150), ttl = self.cashe.get("ttl", 300))
        self.session = await http.AioSession.creat_session()
        
        await git.ImageCache.set_assets_dowload(self.asset_save)
        
        if self.character_id:
            self.character_id = await options.get_charter_id(self.character_id)
        
        if self.character_art:
            if not isinstance(self.character_art, dict):
                raise TypeError(4,"The character_art parameter must be a dictionary, where the key is the name of the character, and the parameter is an image.\nExample: character_art = {'1235': 'img.png', '1235': ['img.png','http.../img2.png']} or {'123596': 'img.png', '123854': 'http.../img2.png', ...}")
            else:
                self.character_art = await options.get_character_art(self.character_art)
        
        if not self.lang in translator.SUPPORTED_LANGUAGES:
            self.lang = "en"
        
        if self.lang == "ua":
            await ukrainization.TranslateDataManager().check_update()
        
        self.translateLang = translator.Translator(self.lang)

            
        if self.user_font:
            await git.change_font(font_path = self.user_font)
        
        if self.remove_logo:
            print("""
                Thank you for using our StarRailCard!
                By removing the GitHub logo from the generated results, you acknowledge supporting the author through one of the following links:
                - Patreon: https://www.patreon.com/deviantapi/membership
                - Ko-fi: https://ko-fi.com/dezzso

                Failure to contribute to the project may lead to the author discontinuing updates in the future.
                """)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.session.close()
    
    async def set_lang(self, lang):
        """Sets the language

        Args:
            lang (str): lang (str): Language in which generation will take place, supported languages: cht, cn, de, en, es, fr, id, jp, kr, pt, ru, th, vi, ua . Defaults to "en".
        """
        if not lang in translator.SUPPORTED_LANGUAGES:
            lang = "en"
            self.lang = "en"
        else:
            self.lang = lang
        
        if lang == "ua":
            await ukrainization.TranslateDataManager().check_update()
        
        self.translateLang = translator.Translator(lang)
                
    async def set_user_font(self, user_font):
        """Will install a custom font

        Args:
            user_font (srt): Font path or font name.
        """
        await git.change_font(font_path = user_font)
    
    async def creat_profile(self, uid, style = 1, hide_uid = False, background = None, force_update = False):
        """Function for generating a user profile card

        Args:
            uid (int): UID of the user in the game.
            style (int, optional): Card style. Defaults to 1.
            hide_uid (bool, optional): Hide UID. Defaults to False.
            background (str, optional): Link to custom card background. Defaults to None.
            force_update (bool, optional): Forced update of user data. Defaults to False.

        Returns:
            StarRail: A class object containing profile information and a profile card
        """
        data = await api.ApiMiHoMo(uid, lang= self.lang, force_update = force_update).get()
        
        response = {
            "settings": {
                "uid": int(uid),
                "lang": self.lang,
                "hide_uid": hide_uid,
                "save": self.save,
                "force_update": force_update,
                "style": int(style)
            },
            "player": data.player,
            "card": None,
            "character_name": [],
            "character_id": [],
        }
        
        if not str(style) in ["1"]:
            style = 1
            
        if style == 1:
            response["card"] = await style_profile_phone.Creat(data,self.translateLang,self.character_art,hide_uid,uid,background, self.remove_logo).start()

        for key in data.characters:
            response["character_id"].append(key.id)
            response["character_name"].append(key.name)
        
        await options.save_card(uid,response["card"],"profile")
        
        return StarRailCard.StarRail(**response)
        
    async def creat(self, uid, style = 1, hide_uid = False, force_update = False, style_settings = None):
        """Function for generating character cards

        Args:
            uid (int): UID of the user in the game
            style (int, optional): Card style. Defaults to 1.
            hide_uid (bool, optional): Hide UID. Defaults to False.
            force_update (bool, optional): forced update of user data. Defaults to False.
            style_settings (dict, optional): not implemented yet. Defaults to None.

        Returns:
            StarRail: A class object containing profile information and a character cards
        """
        
        data = await api.ApiMiHoMo(uid, lang= self.lang, force_update = force_update).get()
        task = []
                
        style, style_settings = await options.style_setting(style, style_settings)
        
        try:
            player = data.player.model_dump()
        except:
            player = data.player
        
        response = {
            "settings": {
                "uid": int(uid),
                "lang": self.lang,
                "hide_uid": hide_uid,
                "save": self.save,
                "force_update": force_update,
                "style": int(style)
            },
            "player": player,
            "card": None,
            "character_name": [],
            "character_id": [],
        }
        
        for key in data.characters:
            response["character_id"].append(key.id)
            response["character_name"].append(key.name)
            
            if self.character_id:
                if not str(key.id) in self.character_id:
                    continue  
            
            art = None
            if self.character_art:
                if str(key.id) in self.character_art:
                    art = self.character_art[str(key.id)]
                    
            if style == 1:
                task.append(style_relict_score.Creat(key,self.translateLang,art,hide_uid,uid, self.seeleland,self.remove_logo).start())
            elif style == 2:
                task.append(style_ticket.Creat(key,self.translateLang,art,hide_uid,uid, self.seeleland,self.remove_logo).start())
        
        response["card"] = await asyncio.gather(*task)
        
        if self.lang == "ua":
            StarRailCard.UA_LANG = True
        else:
            StarRailCard.UA_LANG = False
        task_save = []
        if self.save:
            for key in response["card"]:
                if key["animation"]:
                    continue
                task_save.append(options.save_card(uid,key["card"],key["id"]))
        
            await asyncio.gather(*task_save)
        
        return StarRailCard.StarRail(**response)


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.
import colorsys
from .. import cashe 

_caches = cashe.Cache.get_cache()

async def apply_opacity(image, opacity=0.2):
    result_image = image.copy()
    alpha = result_image.split()[3]
    alpha = alpha.point(lambda p: int(p * opacity))
    result_image.putalpha(alpha)

    return result_image

async def light_level(pixel_color):
    cache_key = pixel_color
    if cache_key in _caches:
        return _caches[cache_key]
    
    h, l, s = colorsys.rgb_to_hls(*(x / 255 for x in pixel_color[:3])) 
    _caches[cache_key] = l
    return l

def color_distance(color1, color2):
    return sum((a - b) ** 2 for a, b in zip(color1, color2)) ** 0.5

async def replace_color(image, old_color, new_color, radius=100):
    image = image.convert("RGBA")
    pixels = image.load()
    width, height = image.size

    for y in range(height):
        for x in range(width):
            current_color = pixels[x, y][:3]
            if color_distance(current_color, old_color) <= radius:
                pixels[x, y] = (*new_color, pixels[x, y][3])
    
    return image


async def recolor_image(image, target_color, light = False):
    if light:
        ll = await light_level(target_color)
        if ll < 45:
            target_color = await _get_light_pixel_color(target_color,up = True)
    if image.mode != 'RGBA':
        image = image.convert('RGBA')

    image = image.copy()

    pixels = image.load()
    for i in range(image.width):
        for j in range(image.height):
            r, g, b, a = pixels[i, j]
            if a != 0:
                pixels[i, j] = target_color + (a,)
    if light:
        return image, target_color
    return image

async def _get_light_pixel_color(pixel_color, up = False):
    h, l, s = colorsys.rgb_to_hls(*(x / 255 for x in pixel_color[:3]))
    if up:
        l = min(max(0.6, l), 0.9)
    else:
        l = min(max(0.3, l), 0.8)
    return tuple(round(x * 255) for x in colorsys.hls_to_rgb(h, l, s))
  
async def _get_dark_pixel_color(pixel_color):
    h, l, s = colorsys.rgb_to_hls(*(x / 255 for x in pixel_color[:3]))
    l = min(max(0.8, l), 0.2)
    a = tuple(round(x * 255) for x in colorsys.hls_to_rgb(h, l, s))
    
    return  a







# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

from PIL import Image, ImageDraw, ImageChops
import numpy as np
import colorsys

class GrandientBackground:
    def __init__(self, image, size, overlay = None, size_art = None):
        self.image = image
        self.overlay = overlay
        self.size_art = size_art
        self.width = size[0]
        self.height = size[1]

    def get_centered_image(self, size):
        background_image = Image.new('RGBA', size, color=(0, 0, 0, 0))
        foreground_image = self.image

        scale = max(size[0] / foreground_image.size[0], size[1] / foreground_image.size[1])
        foreground_image = foreground_image.resize((int(foreground_image.size[0] * scale), int(foreground_image.size[1] * scale)))

        background_size = background_image.size
        foreground_size = foreground_image.size

        x = background_size[0] // 2 - foreground_size[0] // 2

        if foreground_size[1] > background_size[1]:
            y_offset = max(int(0.3 * (foreground_size[1] - background_size[1])), int(0.5 * (-foreground_size[1])))
            y = -y_offset
        else:
            y = background_size[1] // 2 - foreground_size[1] // 2

        background_image.alpha_composite(foreground_image, (x, y))

        return background_image

    def start(self, art_add = False, overlay_add = False, left = False):
        if self.size_art:
            image = self.get_centered_image(self.size_art)
        else:
            image = self.image
        gradient = Image.new('RGBA', (self.width, self.height))
        draw = ImageDraw.Draw(gradient)

        top_color, middle_color, bottom_color = self.get_image_colors(image, left = left)

        for y in range(self.height):
            t = y / (self.height - 1)
            gradient_color = self.interpolate_color(top_color, middle_color, bottom_color, t)
            draw.line([(0, y), (self.width, y)], fill=gradient_color)

        if overlay_add:
            gradient = ImageChops.screen(gradient, self.overlay)
        
        if art_add:
            mask = self.create_transition_mask(image.width, image.height)
            gradient = ImageChops.composite(image, gradient, mask)

        return gradient

    @staticmethod
    def light_level(pixel_color):
        h, l, s = colorsys.rgb_to_hls(*(x / 255 for x in pixel_color[:3]))    
        return l
    
    @staticmethod
    def _get_light_pixel_color(pixel_color, up=False):
        h, l, s = colorsys.rgb_to_hls(*(x / 255 for x in pixel_color[:3]))
        if up:
            l = min(max(l + 0.2, 0), 1)
        else:
            l = min(max(l - 0.2, 0), 1)
        return tuple(round(x * 255) for x in colorsys.hls_to_rgb(h, l, s))
    
    @staticmethod
    def _get_dark_pixel_color(pixel_color):
        h, l, s = colorsys.rgb_to_hls(*(x / 255 for x in pixel_color[:3]))
        l = min(max(0.8, l), 0.2)
        a = tuple(round(x * 255) for x in colorsys.hls_to_rgb(h, l, s))
        
        return  a
    
    
    def get_image_colors(self, image, num_sections=3, left = False):
        width, height = image.size
        section_height = height // num_sections

        colors = []
        for i in range(num_sections):
            if left:
                left_coord = 0
                right_coord = int(width * 0.1)
            else:
                left_coord = int(width * 0.9)
                right_coord = width
            
            top = i * section_height
            bottom = (i + 1) * section_height
            section = image.crop((left_coord, top, right_coord, bottom))
            pixels = section.getdata()
            filtered_pixels = [pixel[:3] for pixel in pixels if pixel[3] > 10]
            if filtered_pixels:
                avg_color = tuple(np.mean(filtered_pixels, axis=0, dtype=int))
            else:
                avg_color = (0, 0, 0)

            ll = self.light_level(avg_color)
            if ll > 0.75:
                avg_color = self._get_light_pixel_color(avg_color)
                ll = self.light_level(avg_color)
            
            colors.append(avg_color)

        return colors

    @staticmethod
    def interpolate_color(start_color, middle_color, end_color, t):
        start_alpha = start_color[3] if len(start_color) > 3 else 255
        middle_alpha = middle_color[3] if len(middle_color) > 3 else 255
        end_alpha = end_color[3] if len(end_color) > 3 else 255

        if t < 0.5:
            t = t * 2
            r = int((1 - t) * start_color[0] + t * middle_color[0])
            g = int((1 - t) * start_color[1] + t * middle_color[1])
            b = int((1 - t) * start_color[2] + t * middle_color[2])
            a = int((1 - t) * start_alpha + t * middle_alpha)
        else:
            t = (t - 0.5) * 2
            r = int((1 - t) * middle_color[0] + t * end_color[0])
            g = int((1 - t) * middle_color[1] + t * end_color[1])
            b = int((1 - t) * middle_color[2] + t * end_color[2])
            a = int((1 - t) * middle_alpha + t * end_alpha)

        return (r, g, b, a)

    @staticmethod
    def create_transition_mask(width, height):
        mask = Image.new('L', (width, height), 255)
        draw = ImageDraw.Draw(mask)
        for x in range(width):
            alpha = int(255 * 2 * (1 - x / width))
            draw.line([(x, 0), (x, height)], fill=alpha)

        return mask


# Copyright 2024 DEViantUa <t.me/deviant_ua>
# All rights reserved.

from PIL import Image

from .color_controle import light_level, _get_light_pixel_color, _get_dark_pixel_color

class GradientGenerator:
    def __init__(self, source_img_path):
        self.source_img = source_img_path
        self.frame = ()
        self.source_width, self.source_height = self.source_img.size

    async def generate(self, width, height, left = False):
        gradient_img = Image.new("RGB", (width, height))
        top_height = height // 3
        bottom_height = height // 3
        center_height = height - top_height - bottom_height
        if left:
            left = 3
            right = 4
        else:
            left = self.source_width - 142
            right = self.source_width - 141
        top_1 = 1
        bottom_1 = top_height - 1
        top_2 = top_height + 1
        bottom_2 = top_height + center_height - 1
        top_3 = top_height + center_height + 1
        bottom_3 = height - 2

        top_color = await self._get_pixel_color(left, top_1, right, bottom_1)
        ll = await light_level(top_color)
        if ll < 45:
            top_color = await _get_light_pixel_color(top_color)
        elif ll > 200:
            top_color = await _get_dark_pixel_color(top_color)

        center_color = await self._get_pixel_color(left, top_2, right, bottom_2)
        
        ll = await light_level(center_color)
        if ll < 45:
            center_color = await _get_light_pixel_color(center_color)
        elif ll > 200:
            center_color = await _get_dark_pixel_color(center_color)

        bottom_color = await self._get_pixel_color(left, top_3, right, bottom_3)
        ll = await light_level(bottom_color)
        if ll < 45:
            bottom_color = await _get_light_pixel_color(bottom_color)
        elif ll > 200:
            bottom_color = await _get_dark_pixel_color(bottom_color)

        for y in range(top_height):
            for x in range(width):
                ratio = y / (top_height - 1)
                gradient_color = self._get_interpolated_color(top_color, center_color, ratio)
                gradient_img.putpixel((x, y), gradient_color)

        for y in range(center_height):
            for x in range(width):
                ratio = y / (center_height - 1)
                gradient_color = self._get_interpolated_color(center_color, bottom_color, ratio)
                gradient_img.putpixel((x, y + top_height), gradient_color)

        for y in range(bottom_height):
            for x in range(width):
                gradient_color = bottom_color
                gradient_img.putpixel((x, y + top_height + center_height), gradient_color)

        return gradient_img.convert("RGBA")

    async def _get_pixel_color(self, left, top, right, bottom):
        cropped_img = self.source_img.crop((left, top, right, bottom))
        resized_img = cropped_img.convert("RGB").resize((1, 1))
        pixel_color = resized_img.getpixel((0, 0))
        
        return pixel_color
    
    def _get_interpolated_color(self, start_color, end_color, ratio):
        return tuple(int(start_color[i] + (end_color[i] - start_color[i]) * ratio) for i in range(3))